/*------------------------------------------------------------------------------

Copyright 2005 CAMBADA Team, All Rights Reserved
DET/IEETA, University of Aveiro, http://www.ieeta.pt/atri/cambada

Application layer process manager

Modified:
2005-04-08 : Created by Paulo Pedreiras (pedreiras@det.ua.pt)

------------------------------------------------------------------------------*/
#ifndef PMAN_H_INCLUDED
#define PMAN_H_INCLUDED

/*
 * Defines
 *
 */
#define PMAN_ACTIVATE_SIG	SIGCONT	// Signal used to activate processes

#define PMAN_ATTACH		1	// PMAN_init option: attach to an existing process table
#define PMAN_NEW		0	// PMAN_init option: initialize a new process table

#define PMAN_CLFREE		0	// Flag for PMAN_close. Release resources. Called by "scheduler" process 
					//    (shared mem, semaphore and kills active processes). 
#define PMAN_CLLEAVE		1	// Flag for PMAN_close. Deattaches from shared mem, only. Called by "client" processes 

#define PMAN_UPD_IMEDIATE	0x01	// QoS update carried out immediately
#define PMAN_UPD_ONNEXTACT	0x02	// QoS update carried out on next activation

#define PMAN_UPD_PERIOD		0x04	// Update period
#define PMAN_UPD_PHASE		0x08	// Update initial phase
#define PMAN_UPD_DEADLINE	0x10	// Update deadline
#define PMAN_UPD_DELAY          0x20    // Update delay


#define PROC_TABLE_SIZE		  20	// Maximum number of processes managed
#define PNAME_LEN		  32	// Maximum length of the process name string

#define PMAN_NOPID		   0	// Process id undefined 
#define PMAN_NOINDEX		  -1	// Process index in PMAN is undefined 

#define PMAN_MAX_PRED              8    // Maximum number of predecessors that a process may have             
#define PMAN_MAX_SUCC              8    // Maximum number of successors that a process may have             

#define PMAN_NOPENDACT             0    // No pending activation on process
#define PMAN_PENDACT               1    // Pending activation on process

/* Process status */
#define PROC_S_EMPTY            0       // Process entry not filled up  
#define PROC_S_ACTIV            1       // Process was activated (but not yet released)
#define PROC_S_READY		2	// Process ready to execute (i.e. active ) flag  
#define PROC_S_IDLE		3	// Process waiting for next instance
#define PROC_S_PEND		4	// Process activated but with unsatisfied precedences  
#define PROC_S_DELAY		5	// Process activated but with delayed execution  

/* TRACE parameters */
#define PMAN_TRACE
#define PMAN_TRACE_SIZE     20000     // Number of events to record during trace ( events/sec = FPS*SUM(1/Period_i)*2 )


/*
 * Data types
 *
 */
#ifdef PMAN_TRACE
typedef struct {
  int pindex;            // Process index within PMAN table
  int etype;             // Event type {[R]elease;[E]pilog}
  struct timeval etime;  // Event time 
} PROC_TRACE_DATA;
#endif

typedef struct {
  char PROC_name[PNAME_LEN+1];  // Process id string
  pid_t PROC_id;                // Process id
  
  /* generic temporal properties */
  int    PROC_period;           // Process period
  int    PROC_phase;            // Process initial phase
  int    PROC_deadline;         // Process deadline (in ticks)
  int    PROC_delay;            // Process delay (in ms)

  /* precedence constraints */
  char   PROC_pred_name[PMAN_MAX_PRED][PNAME_LEN+1]; // Array with the names of the predecessor processes
  int    PROC_succ_index[PMAN_MAX_SUCC];             // Array with the PMAN table indexes of the successors of the process
  unsigned long   PROC_pred_mask;                    // bitmap with the index of all the predecessor processes set  
  unsigned long   PROC_pred_met;                     // bitmap with the index of predecessor processes that had an instance

  /* QoS data */
  int    PROC_qosdata;            // Offset to data structure containing QoS specific data
  char   PROC_qosupdflag;         // Flag to signal if qos data has been updated

  /* Status & statistics data */
  struct timeval PROC_last_start;  // Activation instant (last instance)
  struct timeval PROC_last_finish; // Finish instant (last instance)

  int  PROC_status;            // Process status {PROC_S_READY}
  
  unsigned int PROC_nact;      // Number of activations
  unsigned int PROC_ndm;       // Number of deadline misses
 
} PROC_TYPE;

typedef struct {
  int nprocs;        // Number of active processes registered
  int ticks;         // System "tick" counter
  int (*QoSupd)();   // QoS update function hook
  int (*DdlnExcpt)();// Deadline exception handling hook
  PROC_TYPE proc[PROC_TABLE_SIZE];
#ifdef PMAN_TRACE
  int evt_lastindex;  // Index of last event recorded
  int evt_firstindex; // Index of first event recorded
  PROC_TRACE_DATA evt_trace[PMAN_TRACE_SIZE]; // Trace data
#endif
} PROC_TABLE_TYPE;

/* 
 * Global vars
 */
int pman_sem_id;                   // Semaphore ID (mutual exclusion)
int pman_shmem_id;                 // Shared memory ID
PROC_TABLE_TYPE * p_table;         // Process table

pthread_t awake_thread;            // Delay threads structure
pthread_attr_t awake_thread_attr;  // Delay threads attributes 


/*
 * Function prototypes
 *
 */
#ifdef __cplusplus
extern "C" {
#endif
  
  int PMAN_init(key_t shmem_pman_key, key_t sem_pman_key, void * QoSfun, int QoSdata_sz, int create_flags);
  int PMAN_close(int close_flags);

  int PMAN_procadd(char * p_name, int p_id,int p_period, int p_phase, int p_deadline, void * p_QoSdata, int p_QoSdata_sz);
  int PMAN_procdel(char * p_name);
  
  int PMAN_attach(char * p_name, pid_t p_id);
  int PMAN_deattach(char * p_name);

  int PMAN_prec_add(char *p_name_pred, char *p_name_succ);

  int PMAN_QoSupd(char * p_name, void * QoSdata, int QoSdata_sz, char when_flag);
  int PMAN_msgupd(char * p_name, int p_period, int p_phase, int p_deadline, int p_delay, char which_flag);

#define PMAN_TPupd(p_name, p_period, p_phase, p_deadline, p_delay, which_flag) (PMAN_msgupd(p_name, p_period, p_phase, p_deadline, p_delay, which_flag))

  int PMAN_epilogue(char *p_name);

  int PMAN_query(PROC_TYPE *pdata, int reset_flag);
  int PMAN_print(void);
  int PMAN_print_prec(void);
#ifdef PMAN_TRACE
  int PMAN_trace_evt_set(char *p_name, char trace_evt);
  int PMAN_trace_save(void);
#endif

  int PMAN_tick(void);  
  int PMAN_release(void);

    int linux_sched_fifo(int pindex);

  
#ifdef __cplusplus
}
#endif

#endif
